# encoding: utf-8

from __future__ import unicode_literals

from copy import deepcopy
from functools import partial

import six

from ..compat import LOOKUP_SEP
from ..exceptions import QueryablePropertyError
from ..utils import get_queryable_property, parametrizable_decorator, reset_queryable_property
from .cache_behavior import CLEAR_CACHE
from .mixins import AnnotationMixin, LookupFilterMixin

RESET_METHOD_NAME = 'reset_property'


@six.python_2_unicode_compatible
class QueryableProperty(object):
    """
    Base class for all queryable properties, which are basically simple
    descriptors with some added methods for queryset interaction.
    """

    cached = False  #: Determines if the result of the getter is cached, like Django's cached_property.
    setter_cache_behavior = CLEAR_CACHE  #: Determines what happens if the setter of a cached property is used.
    filter_requires_annotation = False  #: Determines if using the property to filter requires annotating first.

    # Set the attributes of mixin methods to None for easier checks if a
    # property implements them.
    set_value = None
    get_annotation = None
    get_update_kwargs = None

    def __init__(self):
        self.model = None
        self.name = None
        self.setter_cache_behavior = six.get_method_function(self.setter_cache_behavior)

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self

        # Always check for cached values first regardless of the self.cached
        # since values will also be cached through annotations.
        if self._has_cached_value(obj):
            return self._get_cached_value(obj)
        if not self.get_value:
            raise AttributeError('Unreadable queryable property.')
        value = self.get_value(obj)
        if self.cached:
            self._set_cached_value(obj, value)
        return value

    def __set__(self, obj, value):
        if not self.set_value:
            raise AttributeError("Can't set queryable property.")
        return_value = self.set_value(obj, value)
        # If a value is set and the property is set up to cache values or has
        # a current cached value, invoke the configured setter cache behavior.
        if self.cached or self._has_cached_value(obj):
            self.setter_cache_behavior(self, obj, value, return_value)

    def __reduce__(self):
        # Since queryable property instances only make sense in the context of
        # model classes, they can simply be pickled using their model class and
        # name and loaded back from the model class when unpickling. This also
        # saves memory as unpickled properties will be the exact same object as
        # the one on the model class.
        return get_queryable_property, (self.model, self.name)

    def __str__(self):
        return '.'.join((self.model._meta.app_label, self.model._meta.object_name, self.name))

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, six.text_type(self))

    def get_value(self, obj):  # pragma: no cover
        """
        Getter method for the queryable property, which will be called when
        the property is read-accessed.

        :param django.db.models.Model obj: The object on which the property was accessed.
        :return: The getter value.
        """
        raise NotImplementedError()

    def get_filter(self, cls, lookup, value):  # pragma: no cover
        """
        Generate a :class:`django.db.models.Q` object that emulates filtering
        a queryset using this property.

        :param type cls: The model class of which a queryset should be
                         filtered.
        :param str lookup: The lookup to use for the filter (e.g. 'exact',
                           'lt', etc.)
        :param value: The value passed to the filter condition.
        :return: A Q object to filter using this property.
        :rtype: django.db.models.Q
        """
        raise NotImplementedError()

    def contribute_to_class(self, cls, name):
        if LOOKUP_SEP in name:
            raise QueryablePropertyError('The name of a queryable property must not contain the lookup separator "{}".'
                                         .format(LOOKUP_SEP))
        # Store some useful values on model class initialization.
        self.model = self.model or cls
        self.name = self.name or name
        setattr(cls, name, self)  # Re-append the property to the model class
        # If not already set, also add a method to the model class that allows
        # to reset the cached values of queryable properties.
        if not getattr(cls, RESET_METHOD_NAME, None):
            setattr(cls, RESET_METHOD_NAME, reset_queryable_property)

    def _get_cached_value(self, obj):
        """
        Get the cached value for this property from the given object. Requires
        a cached value to be present.

        :param django.db.models.Model obj: The object to get the cached value
                                           from.
        :return: The cached value.
        """
        return obj.__dict__[self.name]

    def _set_cached_value(self, obj, value):
        """
        Set the cached value for this property on the given object.

        :param django.db.models.Model obj: The object to set the cached value
                                           for.
        :param value: The value to cache.
        """
        obj.__dict__[self.name] = value

    def _has_cached_value(self, obj):
        """
        Check if a value for this property is cached on the given object.

        :param django.db.models.Model obj: The object to check for a cached
                                           value.
        :return: True if a value is cached; otherwise False.
        :rtype: bool
        """
        return self.name in obj.__dict__

    def _clear_cached_value(self, obj):
        """
        Clear the cached value for this property on the given object. Does not
        require a cached value to be present and will do nothing if no value is
        cached.

        :param django.db.models.Model obj: The object to clear the cached value
                                           on.
        """
        obj.__dict__.pop(self.name, None)


class queryable_property(QueryableProperty):
    """
    A queryable property that is intended to be used as a decorator.
    """

    # Set the attributes of the default methods to None since the decorator
    # may be used without implementing these methods.
    get_value = None
    get_filter = None

    def __init__(self, getter=None, cached=None):
        """
        Initialize a new queryable property, optionally using the given getter
        method and getter configuration.

        :param function getter: The method to decorate.
        :param cached: Determines if values obtained by the getter should be
                       cached (similar to ``cached_property``). A value of None
                       means no change.
        :type cached: bool | None
        """
        super(queryable_property, self).__init__()
        self.__doc__ = None
        if getter:
            self(getter)
        if cached is not None:
            self.cached = cached

    def __call__(self, getter):
        # Since the initializer may be used as a parametrized decorator, the
        # resulting object will be called to apply the decorator.
        self.get_value = getter
        if getter.__doc__:
            self.__doc__ = getter.__doc__
        return self

    def _extract_function(self, method_or_function):
        """
        Extract the function from the given function or method. Allows to
        decorate either regular functions or e.g. classmethods with the
        decorators of this property.

        :param method_or_function: The decorated method or function.
        :type method_or_function: function | classmethod | staticmethod
        :return: The actual function object.
        """
        return getattr(method_or_function, '__func__', method_or_function)

    def _clone(self, **kwargs):
        """
        Clone this queryable property while overriding attributes. This is
        necessary whenever an additional decorator is used to not mess up in
        inheritance scenarios.

        :param kwargs: Attributes to override.
        :return: A (modified) clone of this queryable property.
        :rtype: queryable_property
        """
        attrs = deepcopy(self.__dict__)
        attrs.update(kwargs)
        clone = self.__class__()
        clone.__dict__.update(attrs)
        return clone

    @parametrizable_decorator
    def getter(self, method, cached=None):
        """
        Decorator for a function or method that is used as the getter of this
        queryable property. May be used as a parameter-less decorator
        (``@getter``) or as a decorator with keyword arguments
        (``@getter(cached=True)``).

        :param function method: The method to decorate.
        :param cached: If True, values returned by the decorated getter method
                       will be cached. A value of None means no change.
        :type cached: bool | None
        :return: A cloned queryable property.
        :rtype: queryable_property
        """
        clone = self._clone()
        if cached is not None:
            clone.cached = cached
        return clone(method)

    @parametrizable_decorator
    def setter(self, method, cache_behavior=None):
        """
        Decorator for a function or method that is used as the setter of this
        queryable property. May be used as a parameter-less decorator
        (``@setter``) or as a decorator with keyword arguments
        (``@setter(cache_behavior=DO_NOTHING)``).

        :param function method: The method to decorate.
        :param cache_behavior: A function that defines how the setter interacts
                               with cached values. A value of None means no
                               change.
        :type cache_behavior: function | None
        :return: A cloned queryable property.
        :rtype: queryable_property
        """
        attrs = dict(set_value=method)
        if cache_behavior:
            attrs['setter_cache_behavior'] = cache_behavior
        return self._clone(**attrs)

    @parametrizable_decorator
    def filter(self, method, requires_annotation=None, lookups=None, boolean=False):
        """
        Decorator for a function or method that is used to generate a filter
        for querysets to emulate filtering by this queryable property. May be
        used as a parameter-less decorator (``@filter``) or as a decorator with
        keyword arguments (``@filter(requires_annotation=False)``). May be used
        to define a one-for-all filter function or a filter function that will
        be called for certain lookups only using the `lookups` argument.

        :param method: The method to decorate.
        :type method: function | classmethod | staticmethod
        :param requires_annotation: True if filtering using this queryable
                                    property requires its annotation to be
                                    applied first; otherwise False. None if
                                    this information should not be changed.
        :type requires_annotation: bool | None
        :param lookups: If given, the decorated function or method will be used
                        for the specified lookup(s) only. Automatically adds
                        the :class:`LookupFilterMixin` to this property if this
                        is used.
        :type lookups: collections.Iterable[str] | None
        :param boolean: If True, the decorated function or method is expected
                        to be a simple boolean filter, which doesn't take the
                        `lookup` and `value` parameters and should always
                        return a `Q` object representing the positive (i.e.
                        `True`) filter case. The decorator will automatically
                        negate the condition if the filter was called with a
                        `False` value.
        :type boolean: bool
        :return: A cloned queryable property.
        :rtype: queryable_property
        """
        method = extracted = self._extract_function(method)
        if boolean:
            if lookups is not None:
                raise QueryablePropertyError('A boolean filter cannot specify lookups at the same time.')

            # Re-use the boolean_filter decorator by simulating a method with
            # a self argument when in reality `method` doesn't have one.
            method = LookupFilterMixin.boolean_filter(lambda prop, model: extracted(model))
            lookups = method._lookups
            method = partial(method, None)

        attrs = {}
        if requires_annotation is not None:
            attrs['filter_requires_annotation'] = requires_annotation
        if lookups is not None:  # Register only for the given lookups.
            attrs['lookup_mappings'] = dict(getattr(self, 'lookup_mappings', {}),
                                            **{lookup: method for lookup in lookups})
        else:  # Register as a one-for-all filter function.
            attrs['get_filter'] = method
        clone = self._clone(**attrs)
        # If the decorated function/method is used for certain lookups only,
        # add the LookupFilterMixin into the new property to be able to reuse
        # its filter implementation based on the lookup mappings.
        if lookups is not None and not isinstance(clone, LookupFilterMixin):
            LookupFilterMixin.inject_into_object(clone)
        return clone

    def annotater(self, method):
        """
        Decorator for a function or method that is used to generate an
        annotation to represent this queryable property in querysets. The
        :class:`AnnotationMixin` will automatically applied to this property
        when this decorator is used.

        :param method: The method to decorate.
        :type method: function | classmethod | staticmethod
        :return: A cloned queryable property.
        :rtype: queryable_property
        """
        clone = self._clone(get_annotation=self._extract_function(method))
        # Dynamically add the AnnotationMixin into the new property to allow
        # to use the default filter implementation. Since an explicitly set
        # filter implementation is stored in the instance dict, it will be used
        # over the default implementation.
        if not isinstance(clone, AnnotationMixin):
            AnnotationMixin.inject_into_object(clone)
        return clone

    def updater(self, method):
        """
        Decorator for a function or method that is used to resolve an update
        keyword argument for this queryable property into the actual update
        keyword arguments.

        :param method: The method to decorate.
        :type method: function | classmethod | staticmethod
        :return: A cloned queryable property.
        :rtype: queryable_property
        """
        return self._clone(get_update_kwargs=self._extract_function(method))



# encoding: utf-8

import operator

from django.db.models import BooleanField

from ..utils import MISSING_OBJECT, ModelAttributeGetter
from .base import QueryableProperty
from .mixins import AnnotationMixin, boolean_filter, LookupFilterMixin


class BooleanProperty(LookupFilterMixin, AnnotationMixin, QueryableProperty):
    """
    Internal base class for common properties that return boolean values.
    """

    filter_requires_annotation = False

    def _get_condition(self):  # pragma: no cover
        """
        Build the query filter condition for this boolean property, which is
        used for both the filter and the annotation implementation.

        :return: The filter condition for this property.
        :rtype: django.db.models.Q
        """
        raise NotImplementedError()

    @boolean_filter
    def get_exact_filter(self, cls):
        return self._get_condition()

    def get_annotation(self, cls):
        from django.db.models import Case, When

        return Case(
            When(self._get_condition(), then=True),
            default=False,
            output_field=BooleanField()
        )


class ValueCheckProperty(BooleanProperty):
    """
    A property that checks if an attribute of a model instance or a related
    object contains a certain value or one of multiple specified values and
    returns a corresponding boolean value.

    Supports queryset filtering and CASE/WHEN-based annotating.
    """

    def __init__(self, attribute_path, *values):
        """
        Initialize a new property that checks for certain field values.

        :param str attribute_path: The name of the attribute to compare
                                   against. May also be a more complex path to
                                   a related attribute using dot-notation (like
                                   with :func:`operator.attrgetter`). If an
                                   intermediate value on the path is None, it
                                   will be treated as "no match" instead of
                                   raising an exception. The behavior is the
                                   same if an intermediate value raises an
                                   ObjectDoesNotExist error.
        :param values: The value(s) to check for.
        """
        self.attribute_getter = ModelAttributeGetter(attribute_path)
        self.values = values
        super(ValueCheckProperty, self).__init__()

    def get_value(self, obj):
        return self.attribute_getter.get_value(obj) in self.values

    def _get_condition(self):
        return self.attribute_getter.build_filter('in', self.values)


class RangeCheckProperty(BooleanProperty):
    """
    A property that checks if a static or dynamic value is contained in a range
    expressed by two field values and returns a corresponding boolean value.

    Supports queryset filtering and CASE/WHEN-based annotating.
    """

    def __init__(self, min_attribute_path, max_attribute_path, value, include_boundaries=True, in_range=True,
                 include_missing=False):
        """
        Initialize a new property that checks if a value is contained in a
        range expressed by two field values.

        :param str min_attribute_path: The name of the attribute to get the
                                       lower boundary from. May also be a more
                                       complex path to a related attribute
                                       using dot-notation (like with
                                       :func:`operator.attrgetter`). If an
                                       intermediate value on the path is None,
                                       it will be treated as a missing value
                                       instead of raising an exception. The
                                       behavior is the same if an intermediate
                                       value raises an ObjectDoesNotExist
                                       error.
        :param str max_attribute_path: The name of the attribute to get the
                                       upper boundary from. The same behavior
                                       as for the lower boundary applies.
        :param value: The value which is tested against the boundary. May be a
                      callable which can be called without any arguments, whose
                      return value will then be used as the test value.
        :param bool include_boundaries: Whether or not the value is considered
                                        a part of the range if it is exactly
                                        equal to one of the boundaries.
        :param bool in_range: Configures whether the property should return
                              `True` if the value is in range (`in_range=True`)
                              or if it is out of the range (`in_range=False`).
                              This also affects the impact of the
                              `include_boundaries` and `include_missing`
                              parameters.
        :param bool include_missing: Whether or not a missing value is
                                     considered a part of the range (see the
                                     description of `min_attribute_path`).
                                     Useful e.g. for nullable fields.
        """
        self.min_attribute_getter = ModelAttributeGetter(min_attribute_path)
        self.max_attribute_getter = ModelAttributeGetter(max_attribute_path)
        self.value = value
        self.include_boundaries = include_boundaries
        self.in_range = in_range
        self.include_missing = include_missing
        super(RangeCheckProperty, self).__init__()

    @property
    def final_value(self):
        value = self.value
        if callable(value):
            value = value()
        return value

    def get_value(self, obj):
        value = self.final_value
        min_value = self.min_attribute_getter.get_value(obj)
        max_value = self.max_attribute_getter.get_value(obj)
        lower_operator = operator.le if self.include_boundaries else operator.lt
        greater_operator = operator.ge if self.include_boundaries else operator.gt
        contained = self.include_missing if min_value in (None, MISSING_OBJECT) else greater_operator(value, min_value)
        contained &= self.include_missing if max_value in (None, MISSING_OBJECT) else lower_operator(value, max_value)
        return not (contained ^ self.in_range)

    def _get_condition(self):
        value = self.final_value
        lower_condition = self.min_attribute_getter.build_filter('lte' if self.include_boundaries else 'lt', value)
        upper_condition = self.max_attribute_getter.build_filter('gte' if self.include_boundaries else 'gt', value)
        if self.include_missing:
            lower_condition |= self.min_attribute_getter.build_filter('isnull', True)
            upper_condition |= self.max_attribute_getter.build_filter('isnull', True)
        if not self.in_range:
            return ~lower_condition | ~upper_condition
        return lower_condition & upper_condition


class AggregateProperty(AnnotationMixin, QueryableProperty):

    def __init__(self, aggregate, cached=False):
        """
        Initialize a new property that gets its value by retrieving an
        aggregated value from the database.

        :param django.db.models.Aggregate aggregate: The aggregate to use to
                                                     determine the value of
                                                     this property.
        :param bool cached: Whether or not this property should use a cached
                            getter. If the property is not cached, the getter
                            will perform the corresponding aggregate query on
                            every access.
        """
        super(AggregateProperty, self).__init__()
        self.aggregate = aggregate
        self.cached = cached

    def get_value(self, obj):
        manager = obj.__class__._base_manager
        return manager.filter(pk=obj.pk).aggregate(**{self.name: self.aggregate})[self.name]

    def get_annotation(self, cls):
        return self.aggregate



# encoding: utf-8

import pytest

from django import VERSION as DJANGO_VERSION

from ..conftest import Concat, Value
from ..models import (ApplicationWithClassBasedProperties, ApplicationWithDecoratorBasedProperties,
                      VersionWithClassBasedProperties, VersionWithDecoratorBasedProperties)

pytestmark = [pytest.mark.django_db, pytest.mark.usefixtures('versions')]


class TestAggregateAnnotations(object):

    @pytest.mark.parametrize('model, order_by, reverse, with_selection', [
        # All parametrizations are expected to yield results ordered by the
        # version count (ASC/DESC depending on the reverse parameter).
        (ApplicationWithClassBasedProperties, ('version_count',), False, False),
        (ApplicationWithDecoratorBasedProperties, ('version_count',), False, False),
        (ApplicationWithClassBasedProperties, ('version_count',), False, True),
        (ApplicationWithDecoratorBasedProperties, ('version_count',), False, True),
        (ApplicationWithClassBasedProperties, ('version_count', '-name'), False, False),
        (ApplicationWithDecoratorBasedProperties, ('version_count', '-name'), False, False),
        (ApplicationWithClassBasedProperties, ('-version_count',), True, False),
        (ApplicationWithDecoratorBasedProperties, ('-version_count',), True, False),
        (ApplicationWithClassBasedProperties, ('-version_count',), True, True),
        (ApplicationWithDecoratorBasedProperties, ('-version_count',), True, True),
        (ApplicationWithClassBasedProperties, ('-version_count', 'name'), True, False),
        (ApplicationWithDecoratorBasedProperties, ('-version_count', 'name'), True, False),
    ])
    def test_single_model(self, model, order_by, reverse, with_selection):
        model.objects.all()[0].versions.all()[0].delete()
        queryset = model.objects.all()
        if with_selection:
            queryset = queryset.select_properties('version_count')
        results = list(queryset.order_by(*order_by))
        assert results == sorted(results, key=lambda application: application.version_count, reverse=reverse)
        # Check that ordering by a property annotation does not lead to a
        # selection of the property annotation.
        assert all(model.version_count._has_cached_value(application) is with_selection for application in results)

    @pytest.mark.parametrize('model, order_by, reverse', [
        # All parametrizations are expected to yield results ordered by the
        # version count (ASC/DESC depending on the reverse parameter).
        (VersionWithClassBasedProperties, ('application__version_count',), False),
        (VersionWithDecoratorBasedProperties, ('application__version_count',), False),
        (VersionWithClassBasedProperties, ('application__version_count', '-application__name'), False),
        (VersionWithDecoratorBasedProperties, ('application__version_count', '-application__name'), False),
        (VersionWithClassBasedProperties, ('-application__version_count',), True),
        (VersionWithDecoratorBasedProperties, ('-application__version_count',), True),
        (VersionWithClassBasedProperties, ('-application__version_count', 'application__name'), True),
        (VersionWithDecoratorBasedProperties, ('-application__version_count', 'application__name'), True),
    ])
    def test_across_relation(self, model, order_by, reverse):
        model.objects.all()[0].delete()  # Create a different version count for the application fixtures.
        results = list(model.objects.order_by(*order_by))
        assert results == sorted(results, key=lambda version: version.application.version_count, reverse=reverse)


@pytest.mark.skipif(DJANGO_VERSION < (1, 8), reason="Expression-based annotations didn't exist before Django 1.8")
class TestExpressionAnnotations(object):

    @pytest.mark.parametrize('model, order_by, reverse, with_selection', [
        # All parametrizations are expected to yield results ordered by the
        # full version (ASC/DESC depending on the reverse parameter).
        (VersionWithClassBasedProperties, ('version',), False, False),
        (VersionWithDecoratorBasedProperties, ('version',), False, False),
        (VersionWithClassBasedProperties, ('version',), False, True),
        (VersionWithDecoratorBasedProperties, ('version',), False, True),
        (VersionWithClassBasedProperties, ('version', '-major'), False, False),
        (VersionWithDecoratorBasedProperties, ('version', '-major'), False, False),
        (VersionWithClassBasedProperties, ('-version',), True, False),
        (VersionWithDecoratorBasedProperties, ('-version',), True, False),
        (VersionWithClassBasedProperties, ('-version',), True, True),
        (VersionWithDecoratorBasedProperties, ('-version',), True, True),
        (VersionWithClassBasedProperties, ('-version', 'minor'), True, False),
        (VersionWithDecoratorBasedProperties, ('-version', 'minor'), True, False),
        (VersionWithClassBasedProperties, (Concat(Value('V'), 'version').asc(),), False, False),
        (VersionWithDecoratorBasedProperties, (Concat(Value('V'), 'version').asc(),), False, False),
        (VersionWithClassBasedProperties, (Concat(Value('V'), 'version').asc(),), False, True),
        (VersionWithDecoratorBasedProperties, (Concat(Value('V'), 'version').asc(),), False, True),
        (VersionWithClassBasedProperties, (Concat(Value('V'), 'version').asc(), '-patch'), False, False),
        (VersionWithDecoratorBasedProperties, (Concat(Value('V'), 'version').asc(), '-patch'), False, False),
        (VersionWithClassBasedProperties, (Concat(Value('V'), 'version').desc(),), True, False),
        (VersionWithDecoratorBasedProperties, (Concat(Value('V'), 'version').desc(),), True, False),
        (VersionWithClassBasedProperties, (Concat(Value('V'), 'version').desc(),), True, True),
        (VersionWithDecoratorBasedProperties, (Concat(Value('V'), 'version').desc(),), True, True),
        (VersionWithClassBasedProperties, (Concat(Value('V'), 'version').desc(), 'major'), True, False),
        (VersionWithDecoratorBasedProperties, (Concat(Value('V'), 'version').desc(), 'major'), True, False),
    ])
    def test_single_model(self, model, order_by, reverse, with_selection):
        queryset = model.objects.all()
        if with_selection:
            queryset = queryset.select_properties('version')
        results = list(queryset.order_by(*order_by))
        assert results == sorted(results, key=lambda version: version.version, reverse=reverse)
        # Check that ordering by a property annotation does not lead to a
        # selection of the property annotation
        assert all(model.version._has_cached_value(version) is with_selection for version in results)

    @pytest.mark.parametrize('model, order_by, expected_names', [
        (ApplicationWithClassBasedProperties, ('versions__version',),
         ['My cool App', 'Another App', 'My cool App', 'Another App']),
        (ApplicationWithDecoratorBasedProperties, ('versions__version',),
         ['My cool App', 'Another App', 'My cool App', 'Another App']),
        (ApplicationWithClassBasedProperties, ('versions__version', '-name'),
         ['My cool App', 'Another App', 'My cool App', 'Another App']),
        (ApplicationWithDecoratorBasedProperties, ('versions__version', '-name'),
         ['My cool App', 'Another App', 'My cool App', 'Another App']),
        (ApplicationWithClassBasedProperties, ('-versions__version',),
         ['Another App', 'My cool App', 'Another App', 'My cool App']),
        (ApplicationWithDecoratorBasedProperties, ('-versions__version',),
         ['Another App', 'My cool App', 'Another App', 'My cool App']),
        (ApplicationWithClassBasedProperties, ('-versions__version', 'name'),
         ['Another App', 'My cool App', 'Another App', 'My cool App']),
        (ApplicationWithDecoratorBasedProperties, ('-versions__version', 'name'),
         ['Another App', 'My cool App', 'Another App', 'My cool App']),
        (ApplicationWithClassBasedProperties, (Concat(Value('V'), 'versions__version').asc(),),
         ['My cool App', 'Another App', 'My cool App', 'Another App']),
        (ApplicationWithDecoratorBasedProperties, (Concat(Value('V'), 'versions__version').asc(),),
         ['My cool App', 'Another App', 'My cool App', 'Another App']),
        (ApplicationWithClassBasedProperties, (Concat(Value('V'), 'versions__version').asc(), '-name'),
         ['My cool App', 'Another App', 'My cool App', 'Another App']),
        (ApplicationWithDecoratorBasedProperties, (Concat(Value('V'), 'versions__version').asc(), '-name'),
         ['My cool App', 'Another App', 'My cool App', 'Another App']),
        (ApplicationWithClassBasedProperties, (Concat(Value('V'), 'versions__version').desc(),),
         ['Another App', 'My cool App', 'Another App', 'My cool App']),
        (ApplicationWithDecoratorBasedProperties, (Concat(Value('V'), 'versions__version').desc(),),
         ['Another App', 'My cool App', 'Another App', 'My cool App']),
        (ApplicationWithClassBasedProperties, (Concat(Value('V'), 'versions__version').desc(), 'name'),
         ['Another App', 'My cool App', 'Another App', 'My cool App']),
        (ApplicationWithDecoratorBasedProperties, (Concat(Value('V'), 'versions__version').desc(), 'name'),
         ['Another App', 'My cool App', 'Another App', 'My cool App']),
    ])
    def test_across_relation(self, model, order_by, expected_names):
        version_model = (VersionWithClassBasedProperties if 'ClassBased' in model.__name__
                         else VersionWithDecoratorBasedProperties)
        # Delete versions to make every version number unique
        version_model.objects.get(version='2.0.0', application__name__contains='cool').delete()
        version_model.objects.get(version='1.3.0', application__name__contains='cool').delete()
        version_model.objects.get(version='1.3.1', application__name__startswith='Another').delete()
        version_model.objects.get(version='1.2.3', application__name__startswith='Another').delete()
        assert [app.name for app in model.objects.order_by(*order_by)] == expected_names


@pytest.mark.skipif(DJANGO_VERSION < (1, 11), reason="Explicit subqueries didn't exist before Django 1.11")
class TestSubqueryAnnotations(object):

    @pytest.mark.parametrize('model, order_by, reverse, with_selection', [
        # All parametrizations are expected to yield results ordered by the
        # full version (ASC/DESC depending on the reverse parameter).
        (ApplicationWithClassBasedProperties, ('highest_version',), False, False),
        (ApplicationWithDecoratorBasedProperties, ('highest_version',), False, False),
        (ApplicationWithClassBasedProperties, ('highest_version',), False, True),
        (ApplicationWithDecoratorBasedProperties, ('highest_version',), False, True),
        (ApplicationWithClassBasedProperties, ('highest_version', '-name'), False, False),
        (ApplicationWithDecoratorBasedProperties, ('highest_version', '-name'), False, False),
        (ApplicationWithClassBasedProperties, ('-highest_version',), True, False),
        (ApplicationWithDecoratorBasedProperties, ('-highest_version',), True, False),
        (ApplicationWithClassBasedProperties, ('-highest_version',), True, True),
        (ApplicationWithDecoratorBasedProperties, ('-highest_version',), True, True),
        (ApplicationWithClassBasedProperties, ('-highest_version', 'name'), True, False),
        (ApplicationWithDecoratorBasedProperties, ('-highest_version', 'name'), True, False),
        (ApplicationWithClassBasedProperties, (Concat(Value('V'), 'highest_version').asc(),), False, False),
        (ApplicationWithDecoratorBasedProperties, (Concat(Value('V'), 'highest_version').asc(),), False, False),
        (ApplicationWithClassBasedProperties, (Concat(Value('V'), 'highest_version').asc(),), False, True),
        (ApplicationWithDecoratorBasedProperties, (Concat(Value('V'), 'highest_version').asc(),), False, True),
        (ApplicationWithClassBasedProperties, (Concat(Value('V'), 'highest_version').asc(), '-name'), False, False),
        (ApplicationWithDecoratorBasedProperties, (Concat(Value('V'), 'highest_version').asc(), '-name'), False, False),
        (ApplicationWithClassBasedProperties, (Concat(Value('V'), 'highest_version').desc(),), True, False),
        (ApplicationWithDecoratorBasedProperties, (Concat(Value('V'), 'highest_version').desc(),), True, False),
        (ApplicationWithClassBasedProperties, (Concat(Value('V'), 'highest_version').desc(),), True, True),
        (ApplicationWithDecoratorBasedProperties, (Concat(Value('V'), 'highest_version').desc(),), True, True),
        (ApplicationWithClassBasedProperties, (Concat(Value('V'), 'highest_version').desc(), 'name'), True, False),
        (ApplicationWithDecoratorBasedProperties, (Concat(Value('V'), 'highest_version').desc(), 'name'), True, False),
    ])
    def test_single_model(self, model, order_by, reverse, with_selection):
        model.objects.all()[0].versions.get(version='2.0.0').delete()
        queryset = model.objects.all()
        if with_selection:
            queryset = queryset.select_properties('highest_version')
        results = list(queryset.order_by(*order_by))
        assert results == sorted(results, key=lambda app: app.highest_version, reverse=reverse)
        # Check that ordering by a property annotation does not lead to a
        # selection of the property annotation
        assert all(model.highest_version._has_cached_value(app) is with_selection for app in results)

    @pytest.mark.parametrize('model, order_by, reverse', [
        # All parametrizations are expected to yield results ordered by the
        # application's highest version and version pk (ASC/DESC depending on
        # the reverse parameter).
        (VersionWithClassBasedProperties, ('application__highest_version', 'pk'), False),
        (VersionWithDecoratorBasedProperties, ('application__highest_version', 'pk'), False),
        (VersionWithClassBasedProperties, ('-application__highest_version', '-pk'), True),
        (VersionWithDecoratorBasedProperties, ('-application__highest_version', '-pk'), True),
        (VersionWithClassBasedProperties, (Concat(Value('V'), 'application__highest_version').asc(), 'pk'), False),
        (VersionWithDecoratorBasedProperties, (Concat(Value('V'), 'application__highest_version').asc(), 'pk'), False),
        (VersionWithClassBasedProperties, (Concat(Value('V'), 'application__highest_version').desc(), '-pk'), True),
        (VersionWithDecoratorBasedProperties, (Concat(Value('V'), 'application__highest_version').desc(), '-pk'), True),
    ])
    def test_across_relation(self, model, order_by, reverse):
        model.objects.filter(version='2.0.0')[0].delete()
        results = list(model.objects.order_by(*order_by))
        assert results == sorted(
            results, key=lambda version: (version.application.highest_version, version.pk), reverse=reverse)



# encoding: utf-8

from copy import deepcopy
from functools import wraps

from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Q
from django.utils.tree import Node

from .compat import LOOKUP_SEP
from .exceptions import QueryablePropertyDoesNotExist

MISSING_OBJECT = object()  #: Arbitrary object to represent that an object in an attribute chain is missing.


def get_queryable_property(model, name):
    """
    Retrieve the :class:`queryable_properties.properties.QueryableProperty`
    object with the given attribute name from the given model class or raise
    an error if no queryable property with that name exists on the model class.

    :param type model: The model class to retrieve the property object from.
    :param str name: The name of the property to retrieve.
    :return: The queryable property.
    :rtype: queryable_properties.properties.QueryableProperty
    """
    from .properties import QueryableProperty

    prop = getattr(model, name, None)
    if not isinstance(prop, QueryableProperty):
        raise QueryablePropertyDoesNotExist("{model} has no queryable property named '{name}'".format(
            model=model.__name__, name=name))
    return prop


get_queryable_property.__safe_for_unpickling__ = True


def reset_queryable_property(obj, name):
    """
    Reset the cached value of the queryable property with the given name on the
    given model instance. Read-accessing the property on this model instance at
    a later point will therefore execute the property's getter again.

    :param django.db.models.Model obj: The model instance to reset the cached
                                       value on.
    :param str name: The name of the queryable property.
    """
    prop = get_queryable_property(obj.__class__, name)
    prop._clear_cached_value(obj)


class InjectableMixin(object):
    """
    A base class for mixin classes that are used to dynamically created classes
    based on a base class and the mixin class.
    """

    # Intentionally use a single cache for all subclasses since it is in no way
    # harmful to use a shared cache.
    _created_classes = {}
    # Class attribute to determine if dynamically built classes should receive
    # a custom __reduce__ implementation so their objects can be pickled.
    _dynamic_pickling = True

    def __init__(self, *args, **kwargs):
        super(InjectableMixin, self).__init__(*args, **kwargs)
        self.init_injected_attrs()

    def init_injected_attrs(self):
        """
        Initialize the attributes this mixin contributes. This method will be
        called during :meth:`__init__` and after the mixin was injected into an
        object.
        """
        pass

    @classmethod
    def mix_with_class(cls, base_class, class_name=None):
        """
        Create a new class based on the given base class and this mixin class.
        The created class will also receive a custom :meth:`__reduce__`
        implementation to make its objects picklable.

        :param type base_class: The base class to mix the mixin into.
        :param str class_name: An optional name for the dynamically created
                               class. If None is supplied (default), the class
                               name of the dynamically created class will be
                               the one of the object's original class.
        :return: The generated class.
        :rtype: type
        """
        class_name = str(class_name or base_class.__name__)
        cache_key = (base_class, cls, class_name)
        created_class = cls._created_classes.get(cache_key)
        if created_class is None:
            attrs = {}
            if cls._dynamic_pickling:
                # Make sure objects of a dynamically created class can be pickled.
                def __reduce__(self):
                    get_state = getattr(self, '__getstate__', lambda: self.__dict__)
                    return _unpickle_injected_object, (base_class, cls, class_name), get_state()
                attrs['__reduce__'] = __reduce__

            created_class = cls._created_classes[cache_key] = type(class_name, (cls, base_class), attrs)
        return created_class

    @classmethod
    def inject_into_object(cls, obj, class_name=None):
        """
        Update the given object's class by dynamically generating a new class
        based on the object's original class and this mixin class and changing
        the given object into an object of this new class.

        :param obj: The object whose class should be changed.
        :param str class_name: An optional name for the dynamically created
                               class. If None is supplied (default), the class
                               name of the dynamically created class will be
                               the one of the object's original class.
        """
        obj.__class__ = cls.mix_with_class(obj.__class__, class_name)
        obj.init_injected_attrs()


# This must be a standalone function for Python 2, where it could not be
# pickled being a static method on the InjectableMixin, even if the underlying
# function had the __safe_for_unpickling__ flag.
def _unpickle_injected_object(base_class, mixin_class, class_name=None):
    """
    Callable for the pickler to unpickle objects of a dynamically created class
    based on the InjectableMixin. It creates the base object from the original
    base class and re-injects the mixin class when unpickling an object.

    :param type base_class: The base class of the pickled object before adding
                            the mixin via injection.
    :param type mixin_class: The :class:`InjectableMixin` subclass that was
                             injected into the pickled object.
    :param str class_name: The class name of the pickled object's dynamically
                           created class.
    :return: The initial unpickled object (before the pickler restores the
             object's state).
    """
    obj = base_class.__new__(base_class, ())
    mixin_class.inject_into_object(obj, class_name)
    return obj


_unpickle_injected_object.__safe_for_unpickling__ = True


class TreeNodeProcessor(object):
    """
    A utility class to simplify working with Django's Node objects.
    """

    def __init__(self, node):
        """
        Initialize a new node processor for the given node.

        :param Node node: The node to process.
        """
        self.node = node

    def check_leaves(self, predicate):
        """
        Check if any leaf of this processor's node matches the given predicate.

        :param function predicate: A function that the leaves of the node will
                                   be tested against. Must take a single
                                   parameter that represents the leaf value.
        :return: True if any leaf matched the predicate; otherwise False.
        :rtype: bool
        """
        for child in self.node.children:
            if isinstance(child, Node) and TreeNodeProcessor(child).check_leaves(predicate):
                return True
            elif not isinstance(child, Node) and predicate(child):
                return True
        return False

    def modify_leaves(self, modifier, copy=True):
        """
        Modify all leaves of this processor's node modifier callable.

        :param function modifier: A callable that will be called for every
                                  encountered actual node value (not subnodes)
                                  with that value as its only parameter. It
                                  must returned the replacement value for the
                                  given value.
        :param bool copy: Whether to create a copy of the original node and
                          modify this copy instead of modifying the original
                          node in place.
        :return: The modified node or node copy.
        :rtype: Node
        """
        if copy:
            return TreeNodeProcessor(deepcopy(self.node)).modify_leaves(modifier, copy=False)

        for index, child in enumerate(self.node.children):
            if isinstance(child, Node):
                TreeNodeProcessor(child).modify_leaves(modifier, copy=False)
            else:
                self.node.children[index] = modifier(child)
        return self.node


def parametrizable_decorator(method):
    """
    A decorator for methods (not regular functions!) who themselves are to be
    used as decorators and are to support both a parameter-less decorator usage
    (``@my_decorator``) as well as parametrized decorator usage
    (``@my_decorator(some_param=5)``). This decorator takes care of making the
    distinction between both usages and returning the correct object.

    :param function method: The decorator method to decorate.
    :return: A wrapper method that will replace the decorated method.
    :rtype: function
    """
    @wraps(method)
    def wrapper(self, function=None, *args, **kwargs):
        def decorator(func):
            return method(self, func, *args, **kwargs)

        if function:  # A function was given directly -> apply the decorator directly (@my_decorator usage).
            return decorator(function)
        return decorator  # No function -> return the actual decorator (@my_decorator(some_param=5) usage).
    return wrapper


class ModelAttributeGetter(object):
    """
    An attribute getter akin to :func:`operator.attrgetter` specifically
    designed for model objects. Like Python's attrgetter, it allows to access
    attributes on related objects using dot-notation, but it catches some
    expected exceptions related to models when following attribute chains.
    It also allows to build filters for querysets based on the configured
    attribute path.
    """

    ATTRIBUTE_SEPARATOR = '.'

    def __init__(self, attribute_path):
        """
        Initialize a new model attribute getter using the specified attribute
        path.

        :param str attribute_path: The path to the attribute to retrieve in
                                   dot-notation (see the docs for
                                   :func:`operator.attrgetter` for examples).
                                   For queryset-related operations, all dots
                                   will simply be replaced with the lookup
                                   separator (``__``).
        """
        self.path_parts = attribute_path.split(self.ATTRIBUTE_SEPARATOR)

    def get_value(self, obj):
        """
        Get the value of the attribute configured in this attribute getter from
        the given model object.

        While resolving the attribute on the model object, a few exceptions are
        automatically caught (leading to the ``MISSING_OBJECT`` constant being
        returned):

        * ``AttributeError``, but only if an object in the attribute chain is
          None. This allows to use this getter in conjunction with nullable
          model fields.
        * ``ObjectDoesNotExist``, which is raised by Django e.g. if an object
          does not exist for reverse one-to-one relations. This allows to use
          this getter in conjunction with such relations as well.

        :param django.db.models.Model obj: The model object to retrieve the
                                           value for.
        :return: The value retrieved from the model object.
        """
        for attribute_name in self.path_parts:
            try:
                obj = getattr(obj, attribute_name)
            except ObjectDoesNotExist:
                # Allow missing DB objects without raising an error, e.g. for
                # reverse one-to-one relations.
                return MISSING_OBJECT
            except AttributeError:
                # Allow objects in between to be None without raising an error,
                # e.g. for nullable fields.
                if obj is None:
                    return MISSING_OBJECT
                raise
        return obj

    def build_filter(self, lookup, value):
        """
        Build a filter condition based on the configured attribute and the
        given lookup and value.

        :param str lookup: The lookup to use for the filter condition.
        :param value: The value to filter against.
        :return: The filter condition as a Q object.
        :rtype: django.db.models.Q
        """
        return Q(**{LOOKUP_SEP.join(self.path_parts + [lookup]): value})



# encoding: utf-8

from datetime import date

from django.db import models

from queryable_properties.managers import QueryablePropertiesManager
from queryable_properties.properties import (
    AggregateProperty, AnnotationMixin, LookupFilterMixin, QueryableProperty, queryable_property, RangeCheckProperty,
    SetterMixin, UpdateMixin, ValueCheckProperty
)


class DummyProperty(SetterMixin, QueryableProperty):

    def __init__(self):
        super(DummyProperty, self).__init__()
        self.counter = 0

    def get_value(self, obj):
        self.counter += 1
        return self.counter

    def set_value(self, obj, value):
        return -1  # A value < 0 to test the CACHE_RETURN_VALUE setter behavior


class HighestVersionProperty(AnnotationMixin, QueryableProperty):

    def get_value(self, obj):
        try:
            return obj.versions.order_by('-major', '-minor', '-patch')[0].version
        except IndexError:
            return None

    def get_annotation(self, cls):
        queryset = VersionWithClassBasedProperties.objects.select_properties('version')
        queryset = queryset.filter(application=models.OuterRef('pk')).order_by('-major', '-minor', '-patch')
        return models.Subquery(queryset.values('version')[:1], output_field=models.CharField())


class VersionCountProperty(AnnotationMixin, QueryableProperty):

    def get_value(self, obj):
        return obj.versions.count()

    def get_annotation(self, cls):
        return models.Count('versions')


class LoweredVersionChangesProperty(AnnotationMixin, QueryableProperty):

    def get_annotation(self, cls):
        from django.db.models.functions import Lower
        return Lower('versions__changes_or_default')


class MajorMinorVersionProperty(UpdateMixin, QueryableProperty):

    def get_value(self, obj):
        return '{major}.{minor}'.format(major=obj.major, minor=obj.minor)

    def get_filter(self, cls, lookup, value):
        if lookup != 'exact':
            raise NotImplementedError()
        parts = value.split('.')
        return models.Q(major=parts[0], minor=parts[1])

    def get_update_kwargs(self, cls, value):
        parts = value.split('.')
        return dict(major=parts[0], minor=parts[1])


class FullVersionProperty(LookupFilterMixin, UpdateMixin, AnnotationMixin, SetterMixin, QueryableProperty):

    filter_requires_annotation = False

    def get_value(self, obj):
        return '{major_minor}.{patch}'.format(major_minor=obj.major_minor, patch=obj.patch)

    def set_value(self, obj, value):
        obj.major, obj.minor, obj.patch = value.split('.')

    @LookupFilterMixin.lookup_filter('exact')
    def exact_filter(self, cls, lookup, value):
        parts = value.rsplit('.', 1)
        return models.Q(major_minor=parts[0], patch=parts[1])

    def get_annotation(self, cls):
        from django.db.models.functions import Concat
        return Concat('major', models.Value('.'), 'minor', models.Value('.'), 'patch', output_field=models.CharField())

    def get_update_kwargs(self, cls, value):
        parts = value.rsplit('.', 1)
        return dict(major_minor=parts[0], patch=parts[1])


class DefaultChangesProperty(AnnotationMixin, QueryableProperty):

    def get_value(self, obj):
        return obj.changes or '(No data)'

    def get_annotation(self, cls):
        from django.db.models import Value
        from django.db.models.functions import Coalesce
        return Coalesce('changes', Value('(No data)'))


class Version2Property(LookupFilterMixin, QueryableProperty):

    def get_value(self, obj):
        return obj.major == 2

    @LookupFilterMixin.boolean_filter
    def exact_filter(self, cls):
        return models.Q(major=2)


class CircularProperty(AnnotationMixin, QueryableProperty):

    def get_annotation(self, cls):
        return models.F('circular')


class Category(models.Model):
    name = models.CharField(max_length=255)

    class Meta:
        abstract = True


class CategoryWithClassBasedProperties(Category):
    objects = QueryablePropertiesManager()

    circular = CircularProperty()

    class Meta:
        verbose_name = 'Category'


class CategoryWithDecoratorBasedProperties(Category):
    objects = QueryablePropertiesManager()

    class Meta:
        verbose_name = 'Category'

    @queryable_property
    def circular(self):
        raise NotImplementedError()

    @circular.annotater
    @classmethod
    def circular(cls):
        return models.F('circular')


class Application(models.Model):
    name = models.CharField(max_length=255)
    common_data = models.IntegerField(default=0)

    class Meta:
        abstract = True


class ApplicationWithClassBasedProperties(Application):
    categories = models.ManyToManyField(CategoryWithClassBasedProperties, related_name='applications')

    objects = QueryablePropertiesManager()

    highest_version = HighestVersionProperty()
    version_count = VersionCountProperty()
    major_sum = AggregateProperty(models.Sum('versions__major'))
    lowered_version_changes = LoweredVersionChangesProperty()
    dummy = DummyProperty()

    class Meta:
        verbose_name = 'Application'


class ApplicationWithDecoratorBasedProperties(Application):
    categories = models.ManyToManyField(CategoryWithDecoratorBasedProperties, related_name='applications')

    objects = QueryablePropertiesManager()

    class Meta:
        verbose_name = 'Application'

    @queryable_property
    def highest_version(self):
        try:
            return self.versions.order_by('-major', '-minor', '-patch')[0].version
        except IndexError:
            return None

    @highest_version.annotater
    @classmethod
    def highest_version(cls):
        queryset = VersionWithDecoratorBasedProperties.objects.select_properties('version')
        queryset = queryset.filter(application=models.OuterRef('pk')).order_by('-major', '-minor', '-patch')
        return models.Subquery(queryset.values('version')[:1], output_field=models.CharField())

    @queryable_property
    def version_count(self):
        return self.versions.count()

    @version_count.annotater
    @classmethod
    def version_count(cls):
        return models.Count('versions')

    @queryable_property
    def major_sum(self):
        return self.versions.aggregate(major_sum=models.Sum('major'))['major_sum']

    @major_sum.annotater
    @classmethod
    def major_sum(cls):
        return models.Sum('versions__major')

    @queryable_property
    def lowered_version_changes(self):
        raise NotImplementedError()

    @lowered_version_changes.annotater
    @classmethod
    def lowered_version_changes(cls):
        from django.db.models.functions import Lower
        return Lower('versions__changes_or_default')


class Version(models.Model):
    ALPHA = 'a'
    BETA = 'b'
    STABLE = 's'
    RELEASE_TYPE_CHOICES = (
        (ALPHA, 'Alpha'),
        (BETA, 'Beta'),
        (STABLE, 'Stable'),
    )

    major = models.IntegerField()
    minor = models.IntegerField()
    patch = models.IntegerField()
    changes = models.TextField(null=True, blank=True)
    release_type = models.CharField(max_length=1, choices=RELEASE_TYPE_CHOICES, default=STABLE)
    supported_from = models.DateField(null=True)
    supported_until = models.DateField(null=True)

    class Meta:
        abstract = True


class VersionWithClassBasedProperties(Version):
    application = models.ForeignKey(ApplicationWithClassBasedProperties, on_delete=models.CASCADE,
                                    related_name='versions')

    objects = QueryablePropertiesManager()

    major_minor = MajorMinorVersionProperty()
    version = FullVersionProperty()
    changes_or_default = DefaultChangesProperty()
    is_version_2 = Version2Property()
    is_alpha = ValueCheckProperty('release_type', Version.ALPHA)
    is_beta = ValueCheckProperty('release_type', Version.BETA)
    is_stable = ValueCheckProperty('release_type', Version.STABLE)
    is_unstable = ValueCheckProperty('release_type', Version.ALPHA, Version.BETA)
    shares_common_data = ValueCheckProperty('application.common_data', 0)
    released_in_2018 = ValueCheckProperty('supported_from.year', 2018)
    is_supported = RangeCheckProperty('supported_from', 'supported_until', date(2019, 1, 1), include_missing=True)
    supported_in_2018 = RangeCheckProperty('supported_from.year', 'supported_until.year', 2018, include_missing=True)

    class Meta:
        verbose_name = 'Version'


class VersionWithDecoratorBasedProperties(Version):
    application = models.ForeignKey(ApplicationWithDecoratorBasedProperties, on_delete=models.CASCADE,
                                    related_name='versions')

    objects = QueryablePropertiesManager()

    class Meta:
        verbose_name = 'Version'

    @queryable_property
    def major_minor(self):
        return '{major}.{minor}'.format(major=self.major, minor=self.minor)

    @major_minor.filter
    @classmethod
    def major_minor(cls, lookup, value):
        if lookup != 'exact':
            raise NotImplementedError()
        parts = value.split('.')
        return models.Q(major=parts[0], minor=parts[1])

    @major_minor.updater
    @classmethod
    def major_minor(cls, value):
        parts = value.split('.')
        return dict(major=parts[0], minor=parts[1])

    @queryable_property
    def version(self):
        return '{major_minor}.{patch}'.format(major_minor=self.major_minor, patch=self.patch)

    @version.setter
    def version(self, value):
        self.major, self.minor, self.patch = value.split('.')

    @version.annotater
    @classmethod
    def version(cls):
        from django.db.models.functions import Concat
        return Concat('major', models.Value('.'), 'minor', models.Value('.'), 'patch', output_field=models.CharField())

    @version.filter(requires_annotation=False, lookups=('exact',))
    @classmethod
    def version(cls, lookup, value):
        parts = value.rsplit('.', 1)
        return models.Q(major_minor=parts[0], patch=parts[1])

    @version.updater
    @classmethod
    def version(cls, value):
        parts = value.rsplit('.', 1)
        return dict(major_minor=parts[0], patch=parts[1])

    @queryable_property
    def changes_or_default(self):
        return self.changes or '(No data)'

    @changes_or_default.annotater
    @classmethod
    def changes_or_default(cls):
        from django.db.models import Value
        from django.db.models.functions import Coalesce
        return Coalesce('changes', Value('(No data)'))

    @queryable_property
    def is_version_2(self):
        return self.major == 2

    @is_version_2.filter(boolean=True)
    @classmethod
    def is_version_2(cls):
        return models.Q(major=2)
